package version2012;
//import java.io.IOException;
import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;


public class Ego implements Cloneable 
{
	public double scorewellbeing;
	public double scoreResource;
	public double scoreAlliance;
	public double scoreRisk;
	public double scoreStress;
	private int id;
	private double eatCount;
	private double meetCount;
	private double fitnessWellBeing;
	private double fitnessResource; 
	private double fitnessRisk; 
	private double fitnessAlliance; 
	private double fitnessStress; 
	private List<MeetLOG> allMeetLog;
	private int interIteration;
	private int strongTies;
	private int weakTies;
	private double trust;
	private double trustWeak;
	private double trustStrong;
	
	private Map<Integer, double[]> trustTrace;
	private boolean isTrustTrace;
	private int currentMetID;
	private boolean metSucceed;
	private String[] actionTrace;
	
	
	//eat ratio
	private double eatRatio;		
	//eat ratio
	private double minDefectRate;
	
	
	
	//policies
	//1. strong ties preferred
	private boolean strongTiesPreferred;
	//2. weak ties preferred
	private boolean weakTiesPreferred;
	//3. stage preferred
	private boolean stagePreferred;
	//4. medium preferred
	private boolean mediumPreferred;
	//5. shuffle preferred
	private boolean shufflePreferred;	
	
	double a;
	static double a_mutation_step = 0.0001;
	
	
	Ego(int id, double a)
	{
		this.id = id;
		this.a = a;
		
		eatCount = 0;
		meetCount = 0;
		
		allMeetLog = new ArrayList<MeetLOG>();
		interIteration = 0;
		
		strongTiesPreferred = false;
		weakTiesPreferred = false;
		stagePreferred = false;
		mediumPreferred = false;
		shufflePreferred = false;	
		
		fitnessResource = 0;
		fitnessWellBeing = 0;
		fitnessRisk = 0;
		fitnessAlliance = 0; 
		fitnessStress = 0; 
		interIteration = 0;
	}	
	
	Ego(int id, Ego ego)
	{
		this.id = id;
		this.a = ego.a;
		eatCount = 0;
		meetCount = 0;
		
		minDefectRate = ego.getDefectRate();
		eatRatio = ego.getEatMeetRatio();	
		
		allMeetLog = new ArrayList<MeetLOG>();
		interIteration = 0;	
		
		strongTiesPreferred = ego.isStrongTiesPreferred();
		weakTiesPreferred = ego.isWeakTiesPreferred();
		stagePreferred = ego.isStagePreferred();
		mediumPreferred = ego.isMediumTiesPreferred();
		shufflePreferred = ego.isShufflePreferred();
		
		fitnessResource = 0;
		fitnessWellBeing = 0;
		fitnessRisk = 0;
		fitnessAlliance = 0; 
		fitnessStress = 0; 
		interIteration = 0;
	}

	public void setEatRatio(double eatRatio)
	{
		this.eatRatio = eatRatio;
	}
	

	public void setMinDefectRate(double minDefectRate)
	{
		this.minDefectRate = minDefectRate;
	}
	
	public void setTrustTrace()
	{
		isTrustTrace = true;
		trustTrace = new HashMap<Integer, double[]>();
		actionTrace = new String[Sociaty.iterations];
	}

	public Ego getEgoByDistance(List<Ego> egos, int i)
	{
		int returnedIndex = i;
		while(returnedIndex < 0)
		{
			returnedIndex += egos.size();
		}
		while(returnedIndex >= egos.size())
		{
			returnedIndex -= egos.size();
		}
		return egos.get(returnedIndex);
	}
	
	public void reset()
	{			
		eatCount = 0;
		meetCount = 0;
		fitnessResource = 0;
		fitnessWellBeing = 0;
		fitnessRisk = 0;
		fitnessAlliance = 0; 
		fitnessStress = 0; 
		fitnessAlliance = 0; 
		fitnessStress = 0; 
		interIteration = 0;
		strongTies = 0;
		weakTies = 0;
		trust = 0;
		fitnessRisk = 0;
		fitnessStress = 0;
		trustStrong = 0;
		trustWeak = 0;
		allMeetLog.clear();
	}
	
	public List<MeetLOG> getAllMeetLog()
	{
		return allMeetLog;
	}
		
	public double getMeetCount()
	{
		return meetCount;
	}
	
	public double getEatCount()
	{
		return eatCount;
	}
	
	public int getInterIteration()
	{
		return interIteration;
	}

	public void setMediumTiesPreferred()
	{
		mediumPreferred = true;
	}
	public void setStrongTiesPreferred()
	{
		strongTiesPreferred = true;
	}
	public void setShufflePreferred()
	{
		shufflePreferred = true;
	}
	public void setWeakTiesPreferred()
	{
		weakTiesPreferred = true;
	}
	public void setStagePreferred()
	{
		stagePreferred = true;
	}
	public boolean isStrongTiesPreferred()
	{
		return strongTiesPreferred;
	}
	public boolean isWeakTiesPreferred()
	{
		return weakTiesPreferred;
	}
	public boolean isStagePreferred()
	{
		return stagePreferred;
	}
	public boolean isMediumTiesPreferred()
	{
		return mediumPreferred;
	}
	public boolean isShufflePreferred()
	{
		return shufflePreferred;
	}
	
	public void runATurn(List<Ego> egoes, int g, int i, int population) throws Exception
	{
		double rand = Math.random();
		if(rand < eatRatio)
		{
			eat();
			metSucceed = false;
			if(isTrustTrace)
			{
				actionTrace[i]="eat";
			}			
		}
		else
		{
			meet(egoes, g, i, population);
			if(isTrustTrace)
			{
				actionTrace[i] = "meet "+currentMetID;
				if(metSucceed)
				{
					actionTrace[i] += " cooperated";
				}
				else
				{
					actionTrace[i] += " rejected";
				}
			}		
		}		
		//calculate fitness and update eat/meet ratio
		calculateFitness();		
		interIteration++;	
		
		//trace trust
		if(isTrustTrace)
		{
			double[] trusts = null;
			MeetLOG log = null;
			int logID;
			for(int j = 0; j < allMeetLog.size(); j++)
			{
				log = allMeetLog.get(j);
				logID = log.getID();				
				Iterator<Integer> it = trustTrace.keySet().iterator();
				Integer trustTraceID = null;
				boolean cont = true;
				while(cont && it.hasNext())
				{
					trustTraceID = it.next();
					if(trustTraceID.intValue() == logID)
					{
						trusts = trustTrace.get(trustTraceID);
						cont = false;
					}
				}
				if(cont)
				{
					trusts = new double[Sociaty.iterations];	
					trustTrace.put(new Integer(logID), trusts);
				}
				trusts[i] = log.getTrust();
			}
		}		
	}	
	
	public void debudPrintTrust() throws IOException
	{
		if(isTrustTrace)
		{
//number of weak and medium ties to output 
int numberWeak = 3;
int numberMedium = 3;
int weakCount = 0;
int mediumCount = 0;

			Integer trustTraceID = null;
			double[] trust = null;
			Iterator<Integer> it = trustTrace.keySet().iterator();
			while(it.hasNext())
			{
				trustTraceID = it.next();
				trust = trustTrace.get(trustTraceID);
				
				if(trust[Sociaty.iterations - 1] > Sociaty.strongTiesThr)
				{
					printOneTrace(trustTraceID, trust);
				}
				
				else if(trust[Sociaty.iterations - 1] > Sociaty.mediumTiesThr )
				{
					if(mediumCount < numberMedium)
					{
						printOneTrace(trustTraceID, trust);
						mediumCount++;			
					}		
				}
				else
				{
					if(weakCount < numberWeak)
					{
						if(mediumCount < numberMedium)
						{
							printOneTrace(trustTraceID, trust);
							weakCount++;			
						}	
					}					
				}

			}
		}
	}
	
	private void printOneTrace(Integer trustTraceID, double[] trust) 
	throws IOException
	{
        //Construct the BufferedWriter object
		String filename = id + "-" + trustTraceID.intValue() + ".txt";
		BufferedWriter bufferedWriter = new BufferedWriter(new FileWriter(filename));
		bufferedWriter.append("ego " + trustTraceID.intValue() + "\n");
		for(int i = 0; i < Sociaty.iterations; i++)
		{
			bufferedWriter.append(trust[i] + "\n");
		}
		bufferedWriter.close();
	}
	
	public double get_total_trust()
	{
		double total_trust = 0.0d;
		for(int i = 0; i < allMeetLog.size(); i++)
		{
			total_trust += allMeetLog.get(i).getTrust();
		}
		return total_trust;
	}
	
	public double getTotalNumberTies()
	{
		return allMeetLog.size();
	}
	
	public double getTotalStrongTrust()
	{
		double st = 0.0d;
		for(int i = 0; i < allMeetLog.size(); i++)
		{
			double t = allMeetLog.get(i).getTrust();
			if(t >= Sociaty.strongTiesThr)
			{
				st += t;
			}
		}
		return st;
	}
	
	public double getWeakTrustWithoutMediumTies()
	{
		return trustWeak;
	}
	
	public double getTotalWeakTrust()
	{
		double wt = 0.0d;
		for(int i = 0; i < allMeetLog.size(); i++)
		{
			double t = allMeetLog.get(i).getTrust();
			if(t < Sociaty.mediumTiesThr)
			{
				wt += t;
			}
		}
		return wt;
	}
	
	public double getTotalMediumTrust()
	{
		double mt = 0.0d;
		for(int i = 0; i < allMeetLog.size(); i++)
		{
			double t = allMeetLog.get(i).getTrust();
			if(t >= Sociaty.mediumTiesThr && t < Sociaty.strongTiesThr)
			{
				mt += t;
			}
		}
		return mt;
	}
	
	public double getTrust()
	{
		return trust;
	}
	
	public double[] getTrustDistribution(int num_girds)
	{
		double t = 100.0 / num_girds;
		double[] trusts = new double[num_girds];
		for(int i = 0; i < allMeetLog.size(); i++)
		{
			MeetLOG log = allMeetLog.get(i);
			trusts[(int)Math.floor(log.getTrust() / t)]++;
		}
		return trusts;
	}

	public void removeLog(int removedID)
	{
		boolean cont = true;
		for(int i = 0; cont && i < allMeetLog.size(); i++)
		{
			MeetLOG log = allMeetLog.get(i);
			if(log.getID() == removedID)
			{
				allMeetLog.remove(log);
				cont = false;
			}
		}		
		
		cont = true;
		calculateFitness();
	}
	
	public void calculateFitness()
	{
		clearZeroTrustTies();
		
		double totalIn = 0.0d;
		double totalOut = 0.0d;
		
		fitnessWellBeing = 0;
		strongTies = 0;
		trust = 0;
		fitnessRisk = 0;
		fitnessStress = 0;
		trustStrong = 0;
		trustWeak = 0;
		double averageTrust = 0;
		int ties = 0;
		if(allMeetLog.size() > 0)
		{
			for(int i = 0; i < allMeetLog.size(); i++)
			{ 
				if(allMeetLog.get(i).getTrust()> 0)
				{
					ties++;
				}
				
				if(allMeetLog.get(i).getTrust() >= Sociaty.strongTiesThr)
				{
					strongTies++;
					trustStrong += allMeetLog.get(i).getTrust();
				}	
				else
				{
					trustWeak += allMeetLog.get(i).getTrust();
				}
				trust += allMeetLog.get(i).getTrust();
//				fitnessRisk += allMeetLog.get(i).getRisk();
				totalIn += allMeetLog.get(i).getIn();
				totalOut += allMeetLog.get(i).getOut();
				fitnessWellBeing += allMeetLog.get(i).getTrust();
			}
			averageTrust = fitnessWellBeing / ties;
			fitnessWellBeing = (fitnessWellBeing * totalIn / totalOut) / ties;
			if(Double.isInfinite(fitnessWellBeing))
			{
				fitnessWellBeing = 0;
			}
			
		}
		weakTies = ties - strongTies;
		fitnessResource = (1 + allMeetLog.size() / 3) * eatCount / 100.0;
		fitnessRisk = averageTrust + 0.005 * meetCount;
		fitnessAlliance = averageTrust + 0.05 * ties;
		fitnessStress = 0.1 * ties;
		clearZeroTrustTies();
		
	}
	
	private void clearZeroTrustTies()
	{
		for(int i = 0; i < allMeetLog.size(); i++)
		{
			if(allMeetLog.get(i).getTrust() <= 0)
			{
				allMeetLog.remove(i);
				i--;
			}
		}
	}

	public int getStrongTies()
	{
		int st = 0;
		for(int i = 0; i < allMeetLog.size(); i++)
		{
			if(allMeetLog.get(i).getTrust() >= Sociaty.strongTiesThr)
			{
				st++;
			}
		}
		return st;
	}
	
	public int getMediumTies()
	{
		int mt = 0;
		for(int i = 0; i < allMeetLog.size(); i++)
		{
			if(allMeetLog.get(i).getTrust() >= Sociaty.mediumTiesThr && allMeetLog.get(i).getTrust() < Sociaty.strongTiesThr )
			{
				mt++;
			}
		}
		return mt;
	}
	
//version 1 without medium ties	
	public int getWeakTiesWithoutMudium()
	{
		return weakTies;
	}
//version 2 with medium ties	
	public int getWeakTies()
	{
		int wt = 0;
		for(int i = 0; i < allMeetLog.size(); i++)
		{
			if(allMeetLog.get(i).getTrust() <= Sociaty.mediumTiesThr)
			{
				wt++;
			}
		}
		return wt;
	}
	private void eat()
	{
		eatCount++;
	}
	
	private void meet_continous(List<Ego> egoes)
	{				
		double totalTrust = 0.0d;
		MeetLOG[] tmpLog_s = new MeetLOG[allMeetLog.size()];			
		for(int i = 0; i < allMeetLog.size(); i++)
		{
			totalTrust += allMeetLog.get(i).getTrust();
			tmpLog_s[i] =  allMeetLog.get(i);
		}
		Arrays.sort(tmpLog_s);
		int selectedID = -1;
		double tmpTrust = 0.0d;
		double rand = Math.random() * totalTrust;
		boolean cont = true;
		for(int i = allMeetLog.size() - 1; cont && i >= 0; i--)
		{	
			tmpTrust += a * tmpLog_s[i].getTrust() + (1.0 - a) * tmpLog_s[allMeetLog.size() - 1 - i].getTrust();
			if(rand <= tmpTrust)
			{
				selectedID = tmpLog_s[i].getID();
				cont = false;
			}
		}		
		tryToMeet(egoes, selectedID);
	}
	private void meetStrong(List<Ego> egoes)
	{				
		double totalTrust = 0.0d;
		MeetLOG[] tmpLog = new MeetLOG[allMeetLog.size()];			
		for(int i = 0; i < allMeetLog.size(); i++)
		{
			totalTrust += allMeetLog.get(i).getTrust();
			tmpLog[i] =  allMeetLog.get(i);
		}
		Arrays.sort(tmpLog);
		int selectedID = -1;
		double tmpTrust = 0.0d;
		double rand = Math.random() * totalTrust;
		boolean cont = true;
		for(int i = allMeetLog.size() - 1; cont && i >= 0; i--)
		{	
			tmpTrust += tmpLog[i].getTrust();
			if(rand <= tmpTrust)
			{
				selectedID = tmpLog[i].getID();
				cont = false;
			}
		}		
		tryToMeet(egoes, selectedID);
	}
	
	private void meetNew(List<Ego> egoes, int iteration, int population)
	{		
		Ego SelectedEgo = null;
		double rand = 0;
		if(population == allMeetLog.size() - 1)
		{
			meetStrong(egoes);
			return;
		}
		else
		{
			rand = Math.random() * (population - allMeetLog.size() - 1);
		}		
		int count = 0;
		boolean cont = true;

		for(int i = 0; cont && i < egoes.size(); i++)
		{

			if(egoes.get(i).getID() != id)
			{
				boolean interCont = true;
				for(int j = 0; interCont && j < allMeetLog.size(); j++)
				{
					if(egoes.get(i).getID() == allMeetLog.get(j).getID())
					{
						interCont = false;
					}
				}
				if(interCont)
				{
					count++;
					if(rand < count)
					{
						SelectedEgo = egoes.get(i);
						cont = false;
					}
				}
			}
		}
		currentMetID = SelectedEgo.getID();
		if(SelectedEgo.cooperate(this))
		{
			addMeetLogOut(SelectedEgo);
			metSucceed = true;
		}
		else
		{
			rejectMeetLog(SelectedEgo);
			metSucceed = false;
		}
	}	
	
	private void meetCompletelyRandomly(List<Ego> egoes)
	{		
		int selectedIndex = (int)Math.floor(Math.random() * egoes.size());
		while(egoes.get(selectedIndex).getID() == id)
		{
			selectedIndex = (int)Math.floor(Math.random() * egoes.size());
		}
		Ego SelectedEgo = egoes.get(selectedIndex);
		currentMetID = SelectedEgo.getID();
		if(SelectedEgo.cooperate(this))
		{
			addMeetLogOut(SelectedEgo);
			metSucceed = true;
		}
		else
		{
			rejectMeetLog(SelectedEgo);
			metSucceed = false;
		}
	}
	
	private void meet(List<Ego> egoes, int generation, int iteration, int population) throws Exception
	{
		if(allMeetLog.size() == 0)
		{
			meetCompletelyRandomly(egoes);		
		}	
		else
		{
			double rand0 = Math.random();
			if((allMeetLog.size() == population - 1))
			{
				
			}
			else if((rand0 > allMeetLog.size() / (double)population) 
					&& (allMeetLog.size() < population - 1))
			{
				meetNew(egoes, iteration, population);			
			}
			else
			{				
				meet_continous(egoes);
			}

		}
	}	

	private void tryToMeet(List<Ego> egoes, int selectedID)
	{
		Ego SelectedEgo = null;
		boolean cont = true;
		for(int i = 0; cont && i < egoes.size(); i++)
		{
			if(egoes.get(i).getID() == selectedID)
			{
				SelectedEgo = egoes.get(i);
				cont = false;
			}
		}	
		currentMetID = SelectedEgo.getID();
		if(SelectedEgo.cooperate(this))
		{
			addMeetLogOut(SelectedEgo);	
			metSucceed = true;
		}
		else
		{
			rejectMeetLog(SelectedEgo);
			metSucceed = false; 
		}	
	}	
		
	private boolean cooperate(Ego focal)
	{
		
		if(Sociaty.DEFENSE)
		{
			//calculate defense probability defence policy and penalty
			double denfenseProability = 0.0d;
			int focalID = focal.getID();
			boolean cont = true;
			for(int i = 0; cont && i< allMeetLog.size(); i++)
			{
				if(allMeetLog.get(i).getID() == focalID)
				{
					denfenseProability = Math.max(minDefectRate,
							allMeetLog.get(i).getDefenseProb());
					cont = false;
				}
			}
			double r = Math.random();
			if(r < denfenseProability)
			{
				if(Sociaty.symmetrical)
				{
					rejectMeetLog(focal);
					if(Sociaty.rewardDefection)
					{
						eatCount++;
					}
				}
				return false;
			}
			else
			{
				meetCount++;
				addMeetLogIn(focal);
				return true;
			}			
		}
		else
		{
			meetCount++;
			addMeetLogIn(focal);
			return true;
		}
	}

	private void addMeetLogIn(Ego ego)
	{
		boolean added = false;
		for(int i = 0; !added && i < allMeetLog.size(); i++)
		{
			if(ego.getID() == allMeetLog.get(i).getID())
			{
				allMeetLog.get(i).increaseTrust();
				allMeetLog.get(i).increaseTrustIn();
				allMeetLog.get(i).addIn();
				added = true;
			}
		}
		if(!added)
		{
			//create a new log to add
			MeetLOG log = new MeetLOG(ego.getID());  
			log.increaseTrust();
			log.increaseTrustIn();
			log.addIn();
			allMeetLog.add(log);
		}
	}
	
	private void addMeetLogOut(Ego ego)
	{
		boolean added = false;
		for(int i = 0; !added && i < allMeetLog.size(); i++)
		{
			if(ego.getID() == allMeetLog.get(i).getID())
			{
				allMeetLog.get(i).increaseTrust();
				allMeetLog.get(i).increaseTrustOut();
				allMeetLog.get(i).addOut();
				added = true;
			}
		}
		if(!added)
		{
			//create a new log to add
			MeetLOG log = new MeetLOG(ego.getID());  
			log.increaseTrust();
			log.increaseTrustOut();
			log.addOut();
			allMeetLog.add(log);
		}
	}
	
	private void rejectMeetLog(Ego ego)
	{
		boolean rejected = false;
		for(int i = 0; !rejected && i < allMeetLog.size(); i++)
		{
			if(ego.getID() == allMeetLog.get(i).getID())
			{
				allMeetLog.get(i).decreaseTrust();
				allMeetLog.get(i).decreaseTrustOut();
				allMeetLog.get(i).increaseBeRejected();
				rejected = true;
			}
		}
		if(!rejected)
		{
			//create a new log to add
			MeetLOG log = new MeetLOG(ego.getID());  
			log.decreaseTrust();
			log.decreaseTrustOut();
			log.increaseBeRejected();
			allMeetLog.add(log);
		}
	}

	public void mutate_on_a()
	{
		if(Math.random() > 0.5d)
		{
			a += a_mutation_step;
		}
		else
		{
			a -= a_mutation_step;
		}
	}
	
	public void mutate()
	{
		//mutation on eat/meet ratio
		double r = Math.random();
		if(r < 0.5d)
		{
			int slot = (int)Math.floor(Math.random() * Sociaty.mutationSlots);
			eatRatio  = Sociaty.minEMR + slot * Sociaty.gradEMR;
		}
		
		//mutation on defect rate
		else 
		{
			int slot = (int)Math.floor(Math.random() * Sociaty.mutationSlots);
			minDefectRate  = Sociaty.minD + slot * Sociaty.gradD;
		}
		
		//mutation on strategy
//		else
//		{
//			if(stagePreferred)
//			{
//				stagePreferred = false;
//				double rand = Math.random();
//				if(rand < 0.3333)
//				{					
//					strongTiesPreferred = true;
//				}
//				else if(rand < 0.6666)
//				{
//					weakTiesPreferred = true;
//				}
//				else
//				{
//					mediumPreferred = true;
//				}
//			}
//			else if(strongTiesPreferred)
//			{
//				strongTiesPreferred = false;
//				double rand = Math.random();
//				if(rand < 0.3333)
//				{					
//					stagePreferred = true;
//				}
//				else if(rand < 0.6666)
//				{
//					weakTiesPreferred = true;
//				}
//				else
//				{
//					mediumPreferred = true;
//				}				
//			}
//			else if(weakTiesPreferred)
//			{
//				weakTiesPreferred = false;
//				double rand = Math.random();
//				if(rand < 0.3333)
//				{					
//					stagePreferred = true;
//				}
//				else if(rand < 0.6666)
//				{
//					strongTiesPreferred = true;
//				}
//				else
//				{
//					mediumPreferred = true;
//				}				
//			}
//			else if(mediumPreferred)
//			{
//				mediumPreferred = false;
//				double rand = Math.random();
//				if(rand < 0.3333)
//				{					
//					stagePreferred = true;
//				}
//				else if(rand < 0.6666)
//				{
//					weakTiesPreferred = true;
//				}
//				else
//				{
//					strongTiesPreferred = true;
//				}				
//			}
//		}		
	}
	
	public void tiesDecay()
	{
		for(int i = 0; i < allMeetLog.size(); i++)
		{
			allMeetLog.get(i).waning();
		}
		clearZeroTrustTies();
	}
	
	public int getID()
	{
		return id;
	}

	public double getEatMeetRatio()
	{
		return eatRatio;
	}
	
	public double getDefectRate()
	{
		return minDefectRate;
	}
	
	public double getResource()
	{
		return fitnessResource;
	}

	public double getWellbeing()
	{
		return fitnessWellBeing;
	}

	public double getAlliance()
	{
		return fitnessAlliance;
	}	

	public double getRisk()
	{
		return fitnessRisk;
	}
	
	public double getStress()
	{
		return fitnessStress;
	}
	
	public double getScore(Sociaty sociaty)
	{
		return sociaty.rescouseWeight * scoreResource
			+ sociaty.wellBeingWeight * scorewellbeing
			+ sociaty.allianceWeight * scoreAlliance
			+ sociaty.riskWeight * scoreRisk
			+ sociaty.stressWeight * scoreStress;
	}	
}
